"""Handles errors recieved from the Jiachang API."""
import json
import xmltodict


class JcException(Exception):
    """Base error for jcapi."""

    def __init__(self, message):
        self.message = message


class NotFound(JcException):
    """Raised when a 404 response is recieved from the Jiachang API.
    Occurs when the requested controller, etc. could not be found."""


class Unauthorized(JcException):
    """Raised when unauthorized, but no other recognized details are provided.
    Occurs when token is invalid."""


class BadCredentials(Unauthorized):
    """Raised when provided credentials are incorrect."""


class BadToken(Unauthorized):
    """Raised when director bearer token is invalid."""


ERROR_CODES = {"401": Unauthorized, "404": NotFound}

ERROR_DETAILS = {
    "Permission denied Bad credentials": BadCredentials,
}

DIRECTOR_ERRORS = {
    "Unauthorized": Unauthorized,
}

DIRECTOR_ERROR_DETAILS = {"Expired or invalid token": BadToken}


async def __checkResponseFormat(response_text: str):
    """Known Jiachang authentication API error message formats:
    ```json
    {
        "JcErrorResponse": {
            "code": 401,
            "details": "Permission denied Bad credentials",
            "message": "Permission denied",
            "subCode": 0
        }
    }
    ```
    ```json
    {
        "code": 404,
        "details": "Account with id:000000 not found in DB",
        "message": "Account not found",
        "subCode": 0
    }```
    ```xml
    <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    <JcErrorResponse>
        <code>401</code>
        <details></details>
        <message>Permission denied</message>
        <subCode>0</subCode>
    </JcErrorResponse>
    ```
    Known Jiachang director error message formats:
    ```json
    {
        "error": "Unauthorized",
        "details": "Expired or invalid token"
    }
    ```
    """
    if response_text.startswith("<"):
        return "XML"
    return "JSON"


async def checkResponseForError(response_text: str):
    """Checks a string response from the Jiachang API for error codes.

    Parameters:
        `response_text` - JSON or XML response from Jiachang, as a string.
    """
    if await __checkResponseFormat(response_text) == "JSON":
        dictionary = json.loads(response_text)
    elif await __checkResponseFormat(response_text) == "XML":
        dictionary = xmltodict.parse(response_text)
    if "JcErrorResponse" in dictionary:
        if dictionary["JcErrorResponse"]["details"] in ERROR_DETAILS:
            exception = ERROR_DETAILS.get(dictionary["JcErrorResponse"]["details"])
            raise exception(response_text)
        else:
            exception = ERROR_CODES.get(
                str(dictionary["JcErrorResponse"]["code"]), JcException
            )
            raise exception(response_text)
    elif "code" in dictionary:
        if dictionary["details"] in ERROR_DETAILS:
            exception = ERROR_DETAILS.get(dictionary["details"])
            raise exception(response_text)
        else:
            exception = ERROR_CODES.get(str(dictionary["code"]), JcException)
            raise exception(response_text)
    elif "error" in dictionary:
        if dictionary["details"] in DIRECTOR_ERROR_DETAILS:
            exception = DIRECTOR_ERROR_DETAILS.get(dictionary["details"])
            raise exception(response_text)
        else:
            exception = DIRECTOR_ERRORS.get(str(dictionary["error"]), JcException)
            raise exception(response_text)



"""Controls Jiachang Switch devices.
"""


class JcSwitch:
    def __init__(self, JcDirector, item_id):
        """Creates a Jiachang Switch object.

        Parameters:
            `JcDirector` - A `jcapi.director.JcDirector` object that corresponds to the Jiachang Director that the switch is connected to.

            `item_id` - The Jiachang item ID of the switch.
        Switch varis:
            {
            "type": "switch",
            "attr": {
                "ID": "106",
                "DEVID": "76",
                "NAME": "2015",
                "SYSNAME": "kg",
                "ICON": "3k3",
                "YYBM": null,
                "INUSE": "1",
                "CANDEL": "0",
                "ISR": "1",
                "ISS": "0",
                "ISC": "1",
                "ATTRINT": "0"
            },
            "value": 0,
            "other": null,
            "detail": true,
            "phydev": "1"
        }
        """
        self.director = JcDirector
        self.item_id = item_id

    async def get_state(self):
        """Returns the power state of a dimmer or switch as a boolean (True=on, False=off).
        """
        value = await self.director.get_item_variable_value(self.item_id, var_name=None, dev_type="switch")
        if str(value) == "0":
            return False
        else:
            return True

    async def turn_on(self):
        """turns on a switch.
        """
        data = {
            "rs": "execAttr",
            "rsargs[]": self.item_id,
            "rsargs[m]": "1",
        }
        return await self.director.request(uri="/devattr/devattr.php", params=data)

    async def turn_off(self):
        """turns off a switch.
        """
        data = {
            "rs": "execAttr",
            "rsargs[]": self.item_id,
            "rsargs[m]": "0",
        }
        return await self.director.request(uri="/devattr/devattr.php", params=data)


"""Authenticates with the Jiachang API, retrieves account and registered
controller info, and retrieves a bearer token for connecting to a Jiachang Director.
"""
import aiohttp
import async_timeout
import json
import logging
import datetime

from .error_handling import checkResponseForError

AUTHENTICATION_ENDPOINT = "https://c.jia.mn/App/a/frame/login.php"
GET_CONTROLLERS_ENDPOINT = "https://c.jia.mn/App/a/frame/setLogin.php"

_LOGGER = logging.getLogger(__name__)


class JcAccount:
    def __init__(
        self, username, password, session: aiohttp.ClientSession = None,
    ):
        """Creates a Jiachang account object.

        Parameters:
            `username` - Jiachang account username/phone_No.

            `password` - Jiachang account password.

            `session` - (Optional) Allows the use of an `aiohttp.ClientSession` object for all network requests. This session will not be closed by the library.
            If not provided, the library will open and close its own `ClientSession`s as needed.
        """
        self.username = username
        self.password = password
        self.session = session

    async def __sendAccountAuthRequest(self):
        """Used internally to retrieve an account bearer token. Returns the entire
        JSON response from the Jiachang auth API.
        """
        dataDictionary = {
            "rs": "login",
            "rsargs[0]": self.username,
            "rsargs[1]": self.password
        }

        if self.session is None:
            async with aiohttp.ClientSession() as session:
                with async_timeout.timeout(10):
                    async with session.post(
                        AUTHENTICATION_ENDPOINT, data=dataDictionary
                    ) as resp:
                        await checkResponseForError(await resp.text())
                        return await resp.text()
        else:
            with async_timeout.timeout(10):
                async with self.session.post(
                    AUTHENTICATION_ENDPOINT, data=dataDictionary
                ) as resp:
                    await checkResponseForError(await resp.text())
                    return await resp.text()

    async def __get_director_control_url(self, uri):
        """Used internally to send GET requests to the Jiachang API,
        authenticated with the account bearer token. Returns the entire JSON
        response from the Jiachang auth API.

        Parameters:
            `uri` - Full URI to send GET request to.
        """
        dataDictionary = {
            "rs": "getdturl"
        }
        # _LOGGER.info("token is %s", self.account_bearer_token)
        try:
            url = uri + "?hictoken=" + self.account_bearer_token
            
        except AttributeError:
            msg = "The account bearer token is missing - was your username/password correct? "
            _LOGGER.error(msg)
            raise
        # _LOGGER.info(url)
        if self.session is None:
            async with aiohttp.ClientSession() as session:
                with async_timeout.timeout(10):
                    async with session.post(url, data=dataDictionary) as resp:
                        await checkResponseForError(await resp.text())
                        return await resp.text()
        else:
            with async_timeout.timeout(10):
                async with self.session.post(url, data=dataDictionary) as resp:
                    # await checkResponseForError(await resp.text())
                    # _LOGGER.info(await resp.text())
                    return await resp.text()

    async def __get_controller_token(self, controller_id):
        """get specified controller token. 获取网关的token
        Parameters:
            `controller_id`: ID of the controller. See `getAccountControllers()` for details.
        """
        try:
            url = AUTHENTICATION_ENDPOINT + "?hictoken=" + self.account_bearer_token
        except AttributeError:
            msg = "The account bearer token is missing - was your username/password correct? "
            _LOGGER.error(msg)
            raise
        json_dict = {
            "rs": "changeHIC",
            "rsargs[]": controller_id
        }
        if self.session is None:
            async with aiohttp.ClientSession() as session:
                with async_timeout.timeout(10):
                    async with session.post(
                        url,
                        data=json_dict,
                    ) as resp:
                        await checkResponseForError(await resp.text())
                        return await resp.text()
        else:
            with async_timeout.timeout(10):
                async with self.session.post(
                    url,
                    data=json_dict,
                ) as resp:
                    await checkResponseForError(await resp.text())
                    return await resp.text()

    async def get_account_token(self):
        """Gets an account bearer token for making Jiachang online API requests.
        Returns:
            {
                "login":true,
                "info":"1189120558-1589091143-3386663-d2d60108e2-accad487b8",
                "nohic":false,
                "iswebsocket":true,
                "domain":"c.jia.mn"
            }
        """
        data = await self.__sendAccountAuthRequest()
        json_dict = json.loads(data)
        try:
            self.account_bearer_token = json_dict["info"]
            return self.account_bearer_token
        except KeyError:
            msg = "Did not recieve an account bearer token. Is your username/password correct? "
            _LOGGER.error(msg + data)
            raise

    async def get_controller_url(self):
        """Returns a dictionary of the information for all controllers registered to an account.
        get server address.获取接口请求地址

        Returns:
            ```
            {
                "a": "http://c.jia.mn/App/a",
                "b": "http://182.61.44.102:8381/App/b",
                "c": "http://c.jia.mn/App/c",
                "s": "182.61.44.102:8385",
                "wss": "182.61.44.102:8418",
                "fromdev": "mainapp",
                "on": true
            }
            ```
        """
        data = await self.__get_director_control_url(GET_CONTROLLERS_ENDPOINT)
        json_dict = json.loads(data)
        # _LOGGER.info(json_dict["b"])# print(jsonDictionary["b"])
        return json_dict["b"]

    async def get_director_token(self, controller_id):
        """Returns a dictionary with a director bearer token for making Jiachang Director API requests, and its expiry time (generally 86400 seconds after current time)
        returns:
            "1189120558-1589243259-560371073-9a35aa5e49-bdb97068cf"

        Parameters:
            `controller_id`: Common name of the controller. See `getAccountControllers()` for details.
        """
        data = await self.__get_controller_token(controller_id)
        token = json.loads(data)
        token_expiration = datetime.datetime.now() + datetime.timedelta(
            seconds=7776000
        )
        # print(f"{token} and {token_expiration}")
        return {
            "token": token,
            "token_expiration": token_expiration,
        }


"""Handles communication with a Jiachang Director, and provides functions for getting details about items on the Director.
"""
import aiohttp
import async_timeout
import json
from .error_handling import checkResponseForError
import logging

_LOGGER = logging.getLogger(__name__)


class JcDirector:
    def __init__(
            self,
            ip,
            director_bearer_token,
            session_no_verify_ssl: aiohttp.ClientSession = None,
    ):
        """Creates a Jiachang Director object.

        Parameters:
            `ip` - The IP address of the Jiachang Director/Controller.
            "/home/status.php&hictoken=2802-1572318277-980006311-18755fdbf8-c525d56258"
            `director_bearer_token` - The bearer token used to authenticate with the Director. See `pyJiachang.account.JcAccount.getDirectorBearerToken` for how to get this.

            `session` - (Optional) Allows the use of an `aiohttp.ClientSession` object for all network requests. This session will not be closed by the library.
            If not provided, the library will open and close its own `ClientSession`s as needed.
        """
        self.base_url = ip
        self.headers = {"Content-Type": "application/x-www-form-urlencoded"}
        self.session = session_no_verify_ssl
        self.token = director_bearer_token

    async def request(self, uri, params, async_variable=True):
        """Sends a POST request to the specified API URI. Used to send commands to the Director.
        Returns the Director's JSON response as a string.

        Parameters:
            `uri` - The API URI to send the request to. Do not include the IP address of the Director.

            `command` - The Jiachang command to send.

            `params` - The parameters of the command, provided as a dictionary.
        """
        url = self.base_url + uri + "?hictoken=" + self.token
        json_dict = params
        _LOGGER.error("base_url is %s", uri)

        if self.session is None:
            async with aiohttp.ClientSession() as session:
                with async_timeout.timeout(10):
                    async with session.post(
                            url, data=json_dict
                    ) as resp:
                        # await checkResponseForError(await resp.text())
                        return await resp.text()
        else:
            with async_timeout.timeout(10):
                async with self.session.post(
                        url, data=json_dict
                ) as resp:
                    # await checkResponseForError(await resp.text())
                    return await resp.text()

    async def get_all_item_info(self):
        """Returns a JSON list of all the items on the Director.
        """
        # res = await self.request(uri="/home/status.php", params={"rs":"getDevListJson"})
        # print(res)
        return await self.request(uri="/home/status.php", params={"rs":"getDevListJson"})

    async def get_item_info(self, item_id):
        """Returns a JSON list of the details of the specified item.

        Parameters:
            `item_id` - The Jiachang item ID.
        """
        data = {
            "rs": "getAttr",
            "rsargs[]": item_id
        }
        return await self.request(uri="/devattr/devattr.php", params=data)

    async def get_item_variable_value(self, item_id, var_name, dev_type):
        """Returns the value of the specified variable for the specified item as a string.

        Parameters:
            `item_id` - The Jiachang item ID.

            `var_name` - The Jiachang variable name.
        """
        data = await self.get_item_info(item_id)

        if data:
            json_dictionary = json.loads(data)
            if dev_type == "switch":
                return json_dictionary["value"]
            if dev_type == "color":
                return json_dictionary["value"].get(var_name, 0)
            else:
                return json_dictionary["value"].get(var_name,0)
        else:
            raise ValueError(
                f"Empty response recieved from Director! The variable {var_name} doesn't seem to exist for item {item_id}."
            )




"""Controls Jiachang Light devices.
"""


class JcLight:
    def __init__(self, JcDirector, item_id):
        """Creates a Jiachang Light object.

        Parameters:
            `JcDirector` - A `jcapi.director.JcDirector` object that corresponds to the Jiachang Director that the light is connected to.

            `item_id` - The Jiachang item ID of the light.
        Light varis:
            {
                "type": "color",
                "value": {
                    "m": -1,
                    "m": 1,
                    "l": 100,
                    "w": 0,
                    "s": 0
                },
                "other": null,
                "detail": {
                    "type": 5,
                    "set": {
                        "s": "1",
                        "m": null
                    },
                    "mode": {
                        "11": "跳舞",
                        "12": "彩色呼吸",
                        "13": "彩色跳变",
                        "14": "七彩变幻",
                        "15": "七彩跳变"
                    }
                },
                "attr": {
                    "ID": "13",
                    "DEVID": "5",
                    "NAME": "小灯",
                    "SYSNAME": "color",
                    "ICON": "cd",
                    "YYBM": null,
                    "INUSE": "1",
                    "CANDEL": "0",
                    "ISR": "1",
                    "ISS": "0",
                    "ISC": "1"
                },
                "phydev": "1"
            }
        """
        self.director = JcDirector
        self.item_id = item_id

    async def get_level(self):
        """Returns the level of a dimming-capable light as an int 0-100.
        Will cause an error if called on a non-dimmer light. Use `getState()` instead.
        """
        value = await self.director.get_item_variable_value(self.item_id, "l",dev_type="color")
        return int(value)

    async def get_state(self):
        """Returns the power state of a dimmer or light as a boolean (True=on, False=off).
        """
        value = await self.director.get_item_variable_value(self.item_id, "m", dev_type="color")
        if value == "-1":
            return False
        else:
            return True

    async def set_level(self, level):
        """Sets the light level of a dimmer or turns on/off a light.
        Any `level > 0` will turn on a light, and `level = 0` will turn off a light.

        Parameters:
            `level` - (int) 0-100
        """
        data = {
            "rs": "execAttr",
            "rsargs[]": self.item_id,
            "rsargs[1][m]": "1",
            "rsargs[1][l]": level,
            "rsargs[1][w]": "0",
            "rsargs[1][s]": "0"
        }
        return await self.director.request(uri="/devattr/devattr.php", params=data)

    async def turn_on(self):
        """turns on a light.
        """
        data = {
            "rs": "execAttr",
            "rsargs[]": self.item_id,
            "rsargs[1][m]": "0",
        }
        return await self.director.request(uri="/devattr/devattr.php", params=data)

    async def turn_off(self):
        """turns off a light.
        """
        data = {
            "rs": "execAttr",
            "rsargs[]": self.item_id,
            "rsargs[1][m]": "-1",
        }
        return await self.director.request(uri="/devattr/devattr.php", params=data)